/*
 * file:        HWTimer.h
 * purpose:     class to abstract hardware Timer/Counters
 * project:     c++ framework for avr controllers
 * author:      Reinhard Mantey
 * copyright:   2003-2016 (c) Creative commons - by-nc-sa - some rights reserved
 *              see http://creativecommons.org/licenses/by-nc-sa/3.0/ for details
 */
#ifndef HWTIMER_H
#define HWTIMER_H
#include "halConfig.h"

namespace srd {

enum class TimerMode {
     Normal,
     PWM_PhaseCorrect,
     PWM_PhaseCorrect_8b,
     PWM_PhaseCorrect_9b,
     PWM_PhaseCorrect_10b,
     CTC_OCR,
     FastPWM,
     FastPWM_8b,
     FastPWM_9b,
     FastPWM_10b,
     PWM_PFC_ICR,
     PWM_PFC_OCR,
     PWM_PC_ICR,
     PWM_PC_OCR,
     CTC_ICR,
     FastPWM_ICR,
     FastPWM_OCR
     };


enum class PreScale {
     Stopped,
     PS_1,
     PS_8,
     PS_32,
     PS_64,
     PS_128,
     PS_256,
     PS_1024,
     EXT_T_Falling,
     EXT_T_Rising
     };


// value is the shift amount to place the
// channel mode into config register
enum class Channel {
     A = 6,
     B = 4,
     C = 2
     };


enum class ChannelMode {
     Disconnected,
     Toggle,
     NonInverted,
     Inverted
     };


/**
 * first 8bit hardware timer/counter
 */
class Timer0
{
  static Timer0* mySelf;

public:
  /**
   * constructor to own and start the hardware timer.
   * The first call to construct a timer owns the timer
   * and only an owner may stop and (re-)start a timer.
   */
  Timer0(TimerMode mode, PreScale scale, bool mayBeStopped = true)
   : preScale(scale)
   , mayBeStopped(mayBeStopped) {
    if (!mySelf) {
       mySelf = this;
       setMode(mode);
       if (!mayBeStopped) setPreScale();
       }
    };

  /**
   * start the counter, by setting prescale value
   */
  void start() const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       setPreScale();
       }
    };
  /**
   * stop the counter (clear prescale bits)
   */
  void stop() const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       TIMER0_STOP;
       }
    };
  /**
   * enable pin handling for timer channel (i.e. for pwm usage)
   */
  static void enableChannel(Channel ch, ChannelMode mode) {
    switch (ch) {
      case Channel::A: TIMER0_ENABLE_CH_A(ch, mode); break;
      case Channel::B: TIMER0_ENABLE_CH_B(ch, mode); break;
      default: break;                                      // no channel C on timer 0
      }
    };
  /**
   * compare value may in some modes be the top value
   */
  static void setCompareValue(Channel ch, int16_t cv) {
      switch (ch) {
        case Channel::A: TIMER0_CMPA_SET(cv); break;
        case Channel::B: TIMER0_CMPB_SET(cv); break;
        default: break;
        }
    };

protected:
  static void setMode(TimerMode mode) {
    switch (mode) {
      case TimerMode::Normal:           TIMER0_MODE_NORMAL;       break;
      case TimerMode::PWM_PhaseCorrect: TIMER0_MODE_PWM_PC;       break;
      case TimerMode::CTC_OCR:          TIMER0_MODE_CTC_OCR;      break;
      case TimerMode::FastPWM:          TIMER0_MODE_FAST_PWM;     break;
      case TimerMode::PWM_PC_OCR:       TIMER0_MODE_PWM_PC_OCR;   break;
      case TimerMode::FastPWM_OCR:      TIMER0_MODE_FAST_PWM_OCR; break;
      default:                          TIMER0_STOP;              break;
      }
    };
  void setPreScale() const {
    switch (preScale) {
      case PreScale::PS_1:          TIMER0_PS_1;    break;
      case PreScale::PS_8:          TIMER0_PS_8;    break;
      case PreScale::PS_64:         TIMER0_PS_64;   break;
      case PreScale::PS_256:        TIMER0_PS_256;  break;
      case PreScale::PS_1024:       TIMER0_PS_1024; break;
      case PreScale::EXT_T_Falling: TIMER0_PS_XF;   break;
      case PreScale::EXT_T_Rising:  TIMER0_PS_XR;   break;
      default:                      TIMER0_STOP;    break;
      }
    };

private:
  const PreScale preScale;
  const bool mayBeStopped;
  };


/**
 * second 8bit hardware timer/counter
 */
class Timer2
{
  static Timer2* mySelf;

public:
  /**
   * constructor to own and start the hardware timer.
   * The first call to construct a timer owns the timer
   * and only an owner may stop and (re-)start a timer.
   */
  Timer2(TimerMode mode, PreScale scale, bool mayBeStopped = true)
   : preScale(scale)
   , mayBeStopped(mayBeStopped) {
    if (!mySelf) {
       mySelf = this;
       setMode(mode);
       if (!mayBeStopped) setPreScale();
       }
    };

  /**
   * start the counter, by setting prescale value
   */
  void start() const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       setPreScale();
       }
    };
  /**
   * stop the counter (clear prescale bits)
   */
  void stop() const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       TIMER2_STOP;
       }
    };
  /**
   * enable pin handling for timer channel (i.e. for pwm usage)
   */
  static void enableChannel(Channel ch, ChannelMode mode) {
    switch (ch) {
      case Channel::A: TIMER2_ENABLE_CH_A(ch, mode); break;
      case Channel::B: TIMER2_ENABLE_CH_B(ch, mode); break;
      default: break;                                      // no channel C on timer 0
      }
    };
  /**
   * compare value may in some modes be the top value
   */
  static void setCompareValue(Channel ch, int16_t cv) {
      switch (ch) {
        case Channel::A: TIMER2_CMPA_SET(cv); break;
        case Channel::B: TIMER2_CMPB_SET(cv); break;
        default: break;
        }
    };

protected:
  static void setMode(TimerMode mode) {
    switch (mode) {
      case TimerMode::Normal:           TIMER2_MODE_NORMAL;       break;
      case TimerMode::PWM_PhaseCorrect: TIMER2_MODE_PWM_PC;       break;
      case TimerMode::CTC_OCR:          TIMER2_MODE_CTC_OCR;      break;
      case TimerMode::FastPWM:          TIMER2_MODE_FAST_PWM;     break;
      case TimerMode::PWM_PC_OCR:       TIMER2_MODE_PWM_PC_OCR;   break;
      case TimerMode::FastPWM_OCR:      TIMER2_MODE_FAST_PWM_OCR; break;
      default:                          TIMER2_STOP;              break;
      }
    };
  void setPreScale() const {
    switch (preScale) {
      case PreScale::PS_1:          TIMER2_PS_1;    break;
      case PreScale::PS_8:          TIMER2_PS_8;    break;
      case PreScale::PS_32:         TIMER2_PS_32;   break;
      case PreScale::PS_64:         TIMER2_PS_64;   break;
      case PreScale::PS_128:        TIMER2_PS_128;  break;
      case PreScale::PS_256:        TIMER2_PS_256;  break;
      case PreScale::PS_1024:       TIMER2_PS_1024; break;
      default:                      TIMER2_STOP;    break;
      }
    };

private:
  const PreScale preScale;
  const bool mayBeStopped;
  };


/**
 * first 16bit hardware timer/counter
 */
class Timer1
{
  static Timer1* mySelf;

public:
  /**
   * constructor to own and start the hardware timer.
   * The first call to construct a timer owns the timer
   * and only an owner may stop and (re-)start a timer.
   */
  Timer1(TimerMode mode, PreScale scale, bool mayBeStopped = true)
   : preScale(scale)
   , mayBeStopped(mayBeStopped) {
    if (!mySelf) {
       mySelf = this;
       setMode(mode);
       if (!mayBeStopped) setPreScale();
       }
    };
  /**
   * start the counter, by setting prescale value
   */
  void start() const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       setPreScale();
       }
    };
  /**
   * stop the counter (clear prescale bits)
   */
  void stop() const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       TIMER1_STOP;
       }
    };
  void setICRValue(uint16_t v) const {
    if (mySelf && mySelf == this && mySelf->mayBeStopped) {
       TIMER1_ICR_SET(v);
       }
    };
  static void enableOverflowInterrupt(bool enable = true) {
    if (enable) {
       TIMER1_ENABLE_OVERFLOW;
       }
    else {
       TIMER1_DISABLE_OVERFLOW;
       }
    };
  static void enableChannel(Channel ch, ChannelMode mode) {
    switch (ch) {
      case Channel::A:  TIMER1_ENABLE_CH_A(ch, mode); break;
      case Channel::B:  TIMER1_ENABLE_CH_B(ch, mode); break;
      case Channel::C:  TIMER1_ENABLE_CH_C(ch, mode); break;
      default: break;
      }
    };
  static void setCompareValue(Channel ch, int16_t cv) {
      switch (ch) {
        case Channel::A: TIMER1_CMPA_SET(cv); break;
        case Channel::B: TIMER1_CMPB_SET(cv); break;
        case Channel::C: TIMER1_CMPC_SET(cv); break;
        default: break;
        }
    };

protected:
  static void setMode(TimerMode mode) {
    switch (mode) {
      case TimerMode::Normal:               TIMER1_MODE_NORMAL;       break;
      case TimerMode::PWM_PhaseCorrect_8b:  TIMER1_MODE_PWM_PC_8B;    break;
      case TimerMode::PWM_PhaseCorrect_9b:  TIMER1_MODE_PWM_PC_9B;    break;
      case TimerMode::PWM_PhaseCorrect_10b: TIMER1_MODE_PWM_PC_10B;   break;
      case TimerMode::CTC_OCR:              TIMER1_MODE_CTC_OCR;      break;
      case TimerMode::FastPWM_8b:           TIMER1_MODE_FAST_PWM_8B;  break;
      case TimerMode::FastPWM_9b:           TIMER1_MODE_FAST_PWM_9B;  break;
      case TimerMode::FastPWM_10b:          TIMER1_MODE_FAST_PWM_10B; break;
      case TimerMode::PWM_PC_ICR:           TIMER1_MODE_PWM_PC_ICR;   break;
      case TimerMode::PWM_PC_OCR:           TIMER1_MODE_PWM_PC_OCR;   break;
      case TimerMode::CTC_ICR:              TIMER1_MODE_CTC_ICR;      break;
      case TimerMode::FastPWM_ICR:          TIMER1_MODE_FAST_PWM_ICR; break;
      case TimerMode::FastPWM_OCR:          TIMER1_MODE_FAST_PWM_OCR; break;
      default:                              TIMER1_STOP;              break;
      }
    };
  void setPreScale() const {
    switch (preScale) {
      case PreScale::PS_1:          TIMER1_PS_1;    break;
      case PreScale::PS_8:          TIMER1_PS_8;    break;
      case PreScale::PS_64:         TIMER1_PS_64;   break;
      case PreScale::PS_256:        TIMER1_PS_256;  break;
      case PreScale::PS_1024:       TIMER1_PS_1024; break;
      case PreScale::EXT_T_Falling: TIMER1_PS_XF;   break;
      case PreScale::EXT_T_Rising:  TIMER1_PS_XR;   break;
      default:                      TIMER1_STOP;    break;
      }
    };

private:
  const PreScale preScale;
  const bool mayBeStopped;
  };

}                                                          // namespace srd

#ifdef PLACE_GLOBALS_HERE
srd::Timer0* srd::Timer0::mySelf = nullptr;
srd::Timer1* srd::Timer1::mySelf = nullptr;
srd::Timer2* srd::Timer2::mySelf = nullptr;
#endif
#endif
