/*
 * file:        LCD.h
 * purpose:     class for LCD display handling using only 3 pins
 *              together with a shift register (i.e. 74*164)
 * project:     c++ framework for avr controllers
 * author:      Reinhard Mantey, bit handling is based on the work of Peter Dannegger
 * copyright:   2008-2016 (c) Creative commons - by-nc-sa - some rights reserved
 *              see http://creativecommons.org/licenses/by-nc-sa/3.0/ for details
 */
#include "LCD.h"

#ifdef HAVE_LCD
#include <stdlib.h>

using namespace srd;

// keep in sync with lines from 74*164
const uint8_t LCD::bits[] = { 0x01, 0x08, 0x04, 0x02, 0x20, 0x40, 0x80, 0x10 };
char LCD::scratch[12];

#ifdef LCD_WANT_SAMPLE
# if (LCD_COLS > 20)                 // 123456789-123456789-1234567
  PROGMEM const char LCD::defLine0[] = "LCD Module SRD-27 Pollin x4";
  PROGMEM const char LCD::defLine1[] = "27 Digits/Letters x 2 Lines";
# elif (LCD_COLS < 20)               // 123456789-123456
  PROGMEM const char LCD::defLine0[] = "LCD Module SRD16";
  PROGMEM const char LCD::defLine1[] = "16 Dig. x 2 Line";
# else                               // 123456789-1234567890
  PROGMEM const char LCD::defLine0[] = "LCD Module SRD-20 HP";
  PROGMEM const char LCD::defLine1[] = "20 Digits x 2 Lines.";
# endif
#endif


void LCD::init() {
  this->clock.toOutput();
  this->data.toOutput();
  this->sync.toOutput();

  _delay_ms(16);
  writeCommand(0x38);
  _delay_ms(5);
  writeCommand(0x38);
  _delay_us(110);
  writeCommand(0x38);
  writeCommand(0x38);
  writeCommand(0x08);
  clear();
  writeCommand(0x06);
  writeCommand(0x0C);
  configure(1, 1, 1);                                  // 8bit bus, 2 lines, big font

#ifdef LCD_WANT_SAMPLE
  writeSample();
  _delay_ms(5000);
#endif
  }


#ifdef LCD_WANT_SAMPLE
void LCD::writeSample() {
  printLine(1, defLine0);
  endLine();
  printLine(2, defLine1);
  endLine();
  }
#endif


void LCD::writeByte(uint8_t byte, uint8_t isData) {
  uint8_t i=0;

  // first the serial data
  for (i=0; i < 8; ++i) {
      clock.setLow();
      if (byte & bits[i]) data.setHigh();
      else                data.setLow();
      clock.setHigh();
      }

  if (isData) data.setHigh();
  else        data.setLow();
  sync.setHigh();
  _delay_us(45);
  sync.setLow();
  _delay_us(45);
  }


void LCD::endLine() {
  while (pos < LCD_COLS) writeData(' ');
  }


void LCD::print(char *s) {
  while (*s) writeData(*s++);
  }


void LCD::print(int8_t v) {
  itoa(v, scratch, 10);
  print(scratch);
  writeData(' ');
  }


void LCD::print(uint8_t v) {
  utoa(v, scratch, 10);
  print(scratch);
  writeData(' ');
  }


void LCD::print(int16_t v) {
  itoa(v, scratch, 10);
  print(scratch);
  writeData(' ');
  }


void LCD::print(uint16_t v) {
  utoa(v, scratch, 10);
  print(scratch);
  writeData(' ');
  }


void LCD::print(int32_t v) {
  itoa(v, scratch, 10);
  print(scratch);
  writeData(' ');
  }


void LCD::print(uint32_t v) {
  utoa(v, scratch, 10);
  print(scratch);
  writeData(' ');
  }


void LCD::beginLine(int8_t line) {
  pos = 0;
  switch (line) {
    case 1: writeCommand(0x80); break;
    case 2: writeCommand(0xC0); break;
    case 3: // not yet, need another pin for secondary controller
    case 4:
            break;
    }
  };


void LCD::printLine(uint8_t line, PGM_P s) {
  uint8_t b;
  uint8_t i;

  beginLine(line);
  for (i=0; (b = pgm_read_byte_near(&s[i])); ++i) writeData(b);
  }


void LCD::configure(uint8_t useBits, uint8_t twoLines, uint8_t bigFont) {
  uint8_t cmd = 0x20;

  if (useBits)  cmd |= 0x10;
  if (twoLines) cmd |= 0x08;
  if (bigFont)  cmd |= 0x04;

  writeCommand(cmd);
  writeCommand(0x0C);                                      // display on, cursor off
  writeCommand(0x06);                                      // enable increment mode
  }
#endif
